@@ -2,9 +2,18 @@
 Base Class for DesktopEntry, IconTheme and IconData
 """
 
-import re, os.path, codecs
-from Exceptions import *
+import re, os, stat, io, sys
+from xdg.Exceptions import *
 import xdg.Locale
+from xdg.util import u, PY3
+
+def is_ascii(s):
+    """Return True if a string consists entirely of ASCII characters."""
+    try:
+        s.encode('ascii', 'strict')
+        return True
+    except UnicodeError:
+        return False
 
 class IniFile:
     defaultGroup = ''
@@ -22,7 +31,11 @@ class IniFile:
     def __cmp__(self, other):
         return cmp(self.content, other.content)
 
-    def parse(self, filename, headers):
+    def parse(self, filename, headers=None):
+        '''Parse an INI file.
+        
+        headers -- list of headers the parser will try to select as a default header
+        '''
         # for performance reasons
         content = self.content
 
@@ -30,8 +43,8 @@ class IniFile:
             raise ParsingError("File not found", filename)
 
         try:
-            fd = file(filename, 'r')
-        except IOError, e:
+            fd = io.open(filename, 'r', encoding='utf-8')
+        except IOError as e:
             if debug:
                 raise e
             else:
@@ -55,16 +68,14 @@ class IniFile:
                     content[currentGroup] = {}
             # key
             else:
-                index = line.find("=")
-                key = line[0:index].strip()
-                value = line[index+1:].strip()
+                key, value = line.split("=", 1)
                 try:
                     if debug and self.hasKey(key, currentGroup):
                         raise DuplicateKeyError(key, currentGroup, filename)
                     else:
                         content[currentGroup][key] = value
                 except (IndexError, UnboundLocalError):
-                    raise ParsingError("[%s]-Header missing" % headers[0], filename)
+                    raise ParsingError("Parsing error on key, group missing", filename)
 
         fd.close()
 
@@ -72,12 +83,13 @@ class IniFile:
         self.tainted = False
 
         # check header
-        for header in headers:
-            if content.has_key(header):
-                self.defaultGroup = header
-                break
-        else:
-            raise ParsingError("[%s]-Header missing" % headers[0], filename)
+        if headers:
+            for header in headers:
+                if header in content:
+                    self.defaultGroup = header
+                    break
+            else:
+                raise ParsingError("[%s]-Header missing" % headers[0], filename)
 
     # start stuff to access the keys
     def get(self, key, group=None, locale=False, type="string", list=False):
@@ -86,16 +98,16 @@ class IniFile:
             group = self.defaultGroup
 
         # return key (with locale)
-        if self.content.has_key(group) and self.content[group].has_key(key):
+        if (group in self.content) and (key in self.content[group]):
             if locale:
                 value = self.content[group][self.__addLocale(key, group)]
             else:
                 value = self.content[group][key]
         else:
             if debug:
-                if not self.content.has_key(group):
+                if group not in self.content:
                     raise NoGroupError(group, self.filename)
-                elif not self.content[group].has_key(key):
+                elif key not in self.content[group]:
                     raise NoKeyError(key, group, self.filename)
             else:
                 value = ""
@@ -107,9 +119,7 @@ class IniFile:
             values = [value]
 
         for value in values:
-            if type == "string" and locale == True:
-                value = value.decode("utf-8", "ignore")
-            elif type == "boolean":
+            if type == "boolean":
                 value = self.__getBoolean(value)
             elif type == "integer":
                 try:
@@ -163,8 +173,9 @@ class IniFile:
             group = self.defaultGroup
 
         for lang in xdg.Locale.langs:
-            if self.content[group].has_key(key+'['+lang+']'):
-                return key+'['+lang+']'
+            langkey = "%s[%s]" % (key, lang)
+            if langkey in self.content[group]:
+                return langkey
 
         return key
 
@@ -222,9 +233,11 @@ class IniFile:
         for value in values:
             if type == "string":
                 code = self.checkString(value)
+            if type == "localestring":
+                continue
             elif type == "boolean":
                 code = self.checkBoolean(value)
-            elif type == "number":
+            elif type == "numeric":
                 code = self.checkNumber(value)
             elif type == "integer":
                 code = self.checkInteger(value)
@@ -267,9 +280,7 @@ class IniFile:
             return 1
 
     def checkString(self, value):
-        # convert to ascii
-        if not value.decode("utf-8", "ignore").encode("ascii", 'ignore') == value:
-            return 1
+        return 0 if is_ascii(value) else 1
 
     def checkRegex(self, value):
         try:
@@ -278,7 +289,7 @@ class IniFile:
             return 1
 
     # write support
-    def write(self, filename=None):
+    def write(self, filename=None, trusted=False):
         if not filename and not self.filename:
             raise ParsingError("File not found", "")
 
@@ -290,18 +301,33 @@ class IniFile:
         if os.path.dirname(filename) and not os.path.isdir(os.path.dirname(filename)):
             os.makedirs(os.path.dirname(filename))
 
-        fp = codecs.open(filename, 'w')
-        if self.defaultGroup:
-            fp.write("[%s]\n" % self.defaultGroup)
-            for (key, value) in self.content[self.defaultGroup].items():
-                fp.write("%s=%s\n" % (key, value))
-            fp.write("\n")
-        for (name, group) in self.content.items():
-            if name != self.defaultGroup:
-                fp.write("[%s]\n" % name)
-                for (key, value) in group.items():
-                    fp.write("%s=%s\n" % (key, value))
-                fp.write("\n")
+        with io.open(filename, 'w', encoding='utf-8') as fp:
+
+            # An executable bit signifies that the desktop file is
+            # trusted, but then the file can be executed. Add hashbang to
+            # make sure that the file is opened by something that
+            # understands desktop files.
+            if trusted:
+                fp.write(u("#!/usr/bin/env xdg-open\n"))
+
+            if self.defaultGroup:
+                fp.write(u("[%s]\n") % self.defaultGroup)
+                for (key, value) in self.content[self.defaultGroup].items():
+                    fp.write(u("%s=%s\n") % (key, value))
+                fp.write(u("\n"))
+            for (name, group) in self.content.items():
+                if name != self.defaultGroup:
+                    fp.write(u("[%s]\n") % name)
+                    for (key, value) in group.items():
+                        fp.write(u("%s=%s\n") % (key, value))
+                    fp.write(u("\n"))
+
+        # Add executable bits to the file to show that it's trusted.
+        if trusted:
+            oldmode = os.stat(filename).st_mode
+            mode = oldmode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
+            os.chmod(filename, mode)
+
         self.tainted = False
 
     def set(self, key, value, group=None, locale=False):
@@ -313,10 +339,7 @@ class IniFile:
             key = key + "[" + xdg.Locale.langs[0] + "]"
 
         try:
-            if isinstance(value, unicode):
-                self.content[group][key] = value.encode("utf-8", "ignore")
-            else:
-                self.content[group][key] = value
+            self.content[group][key] = value
         except KeyError:
             raise NoGroupError(group, self.filename)
             
@@ -326,8 +349,6 @@ class IniFile:
         if self.hasGroup(group):
             if debug:
                 raise DuplicateGroupError(group, self.filename)
-            else:
-                pass
         else:
             self.content[group] = {}
             self.tainted = True
@@ -349,15 +370,13 @@ class IniFile:
 
         try:
             if locales:
-                for (name, value) in self.content[group].items():
+                for name in list(self.content[group]):
                     if re.match("^" + key + xdg.Locale.regex + "$", name) and name != key:
-                        value = self.content[group][name]
                         del self.content[group][name]
-            value = self.content[group][key]
-            del self.content[group][key]
+            value = self.content[group].pop(key)
             self.tainted = True
             return value
-        except KeyError, e:
+        except KeyError as e:
             if debug:
                 if e == group:
                     raise NoGroupError(group, self.filename)
@@ -371,20 +390,14 @@ class IniFile:
         return self.content.keys()
 
     def hasGroup(self, group):
-        if self.content.has_key(group):
-            return True
-        else:
-            return False
+        return group in self.content
 
     def hasKey(self, key, group=None):
         # set default group
         if not group:
             group = self.defaultGroup
 
-        if self.content[group].has_key(key):
-            return True
-        else:
-            return False
+        return key in self.content[group]
 
     def getFileName(self):
         return self.filename
